import click

from octohot.cli.regex.colored_texts import Colors
from octohot.cli.github.utils import login, is_archived, check_rate_limit
from octohot.cli.github.pr import _pr
from multiprocessing.pool import ThreadPool


def job(client, c, repo, _click, _opts):
    # Make sure your current rate limit won't exceed
    check_rate_limit(client)

    try:
        if is_archived(repo):
            _click.echo(
                repo.pprint(c.warning('Skipping Repository because it is archived.', bold=True))
            )
            return None

        if repo.tainted():
            _click.echo(
                repo.pprint(c.success("Apply (Pull, change branch, add files, commit, push and make a optional PR)"))
            )
            repo.pull()
            repo.branch(_opts["branch_name"])
            repo.add()
            repo.commit(_opts["commit_name"], _opts["commit_description"])
            repo.push(_opts["branch_name"])
            if _opts["pull_request"] and repo.is_github():
                _pr(client, repo,
                    _opts["commit_name"],
                    _opts["commit_description"],
                    _opts["branch_name"],
                    _opts["default_branch"],
                    _opts["origin_branch"]
                    )
                repo.branch(repo.default_branch())
        else:
            _click.echo(
                repo.pprint(c.warning("nothing to commit, working tree clean"))
            )

        return None

    except Exception as e:
        _click.echo(
            repo.pprint(c.fail(e, bold=True))
        )

    return repo.name


@click.command()
@click.argument('branch_name')
@click.argument('commit_name')
@click.argument('commit_description')
@click.option('--pull_request', '-pr', is_flag=True, default=False)
@click.option('--default-branch', is_flag=True, required=False, default=False)
@click.option('--origin-branch', '-o', multiple=True, required=False)
def apply(branch_name, commit_name, commit_description, pull_request, default_branch, origin_branch):
    """Pull, create branch, add, commit, push and make an optional PR"""
    from octohot.cli.config import repositories
    gh_client = login()
    repo_owner = ''
    color = Colors()
    opts = {
        "branch_name": branch_name,
        "commit_name": commit_name,
        "commit_description": commit_description,
        "pull_request": pull_request,
        "default_branch": default_branch,
        "origin_branch": origin_branch
    }

    # Starting multiple threads
    threads = []
    pool = ThreadPool(processes=4)

    i = 1
    for repo in repositories:
        if not repo_owner:
            repo_owner = repo.owner
        print(color.bold('Starting Job: {} of {}'.format(i, len(repositories))))
        async_result = pool.apply_async(job, (gh_client, color, repo, click, opts))
        threads.append(async_result)
        i += 1

    # Return values based on each thread
    i = 1
    errors = []
    for thread in threads:
        error = thread.get()
        print(f"Completed: {i} of {len(repositories)}")
        if error:
            errors.append(error)
        i += 1

    pool.close()

    if errors:
        print('\nRepositories with Request Errors')
        for err in errors:
            print('- git@github.com:{}/{}'.format(repo_owner, err))



import time
import click
from octohot.cli.github.utils import is_archived
from octohot.cli.regex.colored_texts import Colors
from threading import Thread

color = Colors()


def workflow(repo):
    try:
        if is_archived(repo):
            click.echo(repo.pprint(color.warning('Skipping Repository because it is archived.', bold=True)))
            return
        click.echo(repo.pprint(color.success('Sync: Clone, reset, delete unpushed Branches, pull')))
        repo.clone()
        repo.reset()
        repo.branch(repo.default_branch())
        repo.delete_unpushed_branches()
        repo.pull()
    except Exception as e:
        click.echo(
            repo.pprint(color.fail(e, bold=True))
        )


@click.command()
def sync():
    """Sync: Clone, reset, delete unpushed Branches, pull"""
    from octohot.cli.config import repositories

    threads = []
    i = 0
    count = 0
    for repo in repositories:
        print(color.bold('Repositories: {} of {}'.format(i + 1, len(repositories))))
        worker = Thread(target=workflow, args=(repo,))
        threads.append(worker)
        worker.start()

        count += 1
        if count >= 10:
            count = 0
            time.sleep(10)

        i += 1

    for index, thread in enumerate(threads):
        thread.join()
        print(color.bold('Completed: {} of {}'.format(index + 1, len(threads))))



from os import getenv
from time import sleep
from octohot.cli.config import config
from github3 import GitHub


rate_limit = int(getenv('GITHUB_RATE_LIMIT', "500"))
timeout = int(getenv('GITHUB_TIMEOUT', "10"))


def get_token():
    try:
        token = config['github_token']
    except Exception as e:
        print(e)
        print('\nTrying to use GITHUB_TOKEN environment variable')
        token = getenv('GITHUB_TOKEN', '')

    if not token:
        raise Exception("Missing token")

    return token


def login():
    token = get_token()
    git = GitHub()
    git.login(token=token)
    return git


def is_archived(_repository):
    git = login()
    repo = git.repository(_repository.owner, _repository.name)
    return repo.archived


def check_rate_limit(_client):
    while True:
        remaining = _client.ratelimit_remaining
        if remaining < 500:
            print(f'\n>>>>>> Your rate limit is set to {rate_limit} and you only have {remaining}.')
            print(f'>>>>>> Trying again in {timeout} seconds...\n')
            sleep(timeout)
            continue
        break




class Colors:
    PINK = '\033[95m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    TERMINATE = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

    def warning(self, msg, bold=False, underline=False):
        bold = self.BOLD if bold else ''
        underline = self.UNDERLINE if underline else ''
        return bold + underline + self.YELLOW + str(msg) + self.TERMINATE

    def success(self, msg, bold=False, underline=False):
        bold = self.BOLD if bold else ''
        underline = self.UNDERLINE if underline else ''
        return bold + underline + self.GREEN + str(msg) + self.TERMINATE

    def primary(self, msg, bold=False, underline=False):
        bold = self.BOLD if bold else ''
        underline = self.UNDERLINE if underline else ''
        return bold + underline + self.BLUE + str(msg) + self.TERMINATE

    def fail(self, msg, bold=False, underline=False):
        bold = self.BOLD if bold else ''
        underline = self.UNDERLINE if underline else ''
        return bold + underline + self.RED + str(msg) + self.TERMINATE

    def bold(self, msg):
        return self.BOLD + str(msg) + self.TERMINATE

    def underline(self, msg):
        return self.UNDERLINE + str(msg) + self.TERMINATE

    def header_pink(self, msg):
        return self.PINK + str(msg) + self.TERMINATE



import click

from octohot.cli.regex.colored_texts import Colors


@click.command()
@click.argument('pr_title')
@click.argument('pr_comment')
@click.argument('branch_name')
@click.argument('default_branch')
@click.argument('origin_branch')
def pr(pr_title, pr_comment, branch_name, default_branch, origin_branch):
    from octohot.cli.config import repositories
    for repo in repositories:
        if repo.is_github() and repo.is_current_branch(branch_name):
            _pr(repo, pr_title, pr_comment, branch_name, default_branch, origin_branch)


def _pr(github_login, repo, pr_title, pr_comment, branch_name, default_branch, origin_branch):
    from github3 import exceptions
    color = Colors()
    github_repo = github_login.repository(repo.owner, repo.name)

    origin = 'master'
    if default_branch:
        origin = github_repo.default_branch
    if not default_branch and origin_branch:
        for b in origin_branch:
            try:
                origin = github_repo.branch(b).name
                break
            except exceptions.NotFoundError:
                continue

    pr = github_repo.create_pull(pr_title, origin, branch_name, pr_comment)
    click.echo(color.success("PR github repo %s PR %s branch %s: %s" % (repo.name, pr_title, branch_name, pr.url)))



import click


class File(object):
    def __init__(self, filename=None):
        if not filename:
            raise ValueError("filename not be None")

        self.filename = filename
        self.content = open(self.filename, 'rb').read()
        self.decoded_content = self.content.decode()

    def line(self, charpos):
        return len(self.decoded_content[0:charpos].split('\n'))

    def find(self, pattern):
        import re
        return list(re.finditer(pattern.encode(), self.content))

    def save(self):
        with open(self.filename, 'wb') as f:
            f.write(self.content)

    def replace(self, find, replace, dryrun):
        if dryrun:
            matches = self.find(find)
            for match in matches:
                click.echo("%s: (%s,%s)" % (self.filename, match.start(), match.end()))
        else:
            import re
            self.content = re.sub(find.encode(), replace.replace('\\n', '\n').encode(), self.content)
            self.save()

    def content_by_line(self, line_start, line_end=None):
        if line_end and line_end != line_start:
            out = '\n'.join(self.decoded_content.split('\n')[line_start-1:line_end-1])
        else:
            out = self.decoded_content.split('\n')[line_start-1]

        if len(out) > 80:
            return "%s ..." % out[:80]
        else:
            return out
